﻿using System;
using System.Collections.Generic;
using System.Drawing.Printing;
using System.Management;
using System.Printing;
using System.Printing.Interop;
using System.Reflection;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Markup;
using System.Windows.Media;

namespace Wpf.Print.Sample.Common
{
    public sealed class PrinterManager
    {
        private List<string> ListKeyWord = new List<string> { "mip-201", "mip-001" }; // Driver 이름
        private List<string> ListPrinterName = new List<string> { "nemonic MIP-201", "nemonic MIP-001" }; // Printer 이름(앞일 수록 우선 순위가 높음)
        private string PrinterName = "nemonic MIP-201"; // 기본 Printer 명

        private PrintQueue _printQueue;
        private PrinterSettings _printerSettings;

        private static PrinterManager _instance;
        /// <summary>
        /// PrinterManager 가 여러 번 생성되는 것을 방지하기 위해 Singleton 을 사용합니다
        /// </summary>
        public static PrinterManager Instance
        {
            get
            {
                if (_instance == null)
                    _instance = new PrinterManager();
                return _instance;
            }
        }

        /// <summary>
        /// 프린터가 Online 인지 여부를 가져옵니다
        /// </summary>
        public bool IsReady
        {
            get
            {
                bool isReady = false;

                try
                {
                    foreach (var printerName in ListPrinterName)
                    {
                        // 기존에 저장되어 있는 프린터명으로 네모닉 프린터를 찾습니다
                        var path = $"win32_printer.DeviceId='{printerName}'";

                        using (var managementObject = new ManagementObject(path))
                        {
                            if (managementObject != null)
                            {
                                // https://docs.microsoft.com/ko-kr/windows/win32/cimwin32prov/win32-printer?redirectedfrom=MSDN
                                if (!Convert.ToBoolean(managementObject["WorkOffline"]))
                                {
                                    // 프린터가 Online 일 경우
                                    isReady = true;
                                    PrinterName = printerName;
                                    break;
                                }
                            }
                        }
                    }
                }
                catch (Exception e1)
                { 
                    // 찾는 네모닉 프린터의 프린터명이 nemonic MIP-001 이 아닐 경우 드라이버 이름으로 검색하여 찾습니다
                    // (사용자가 장치 및 프린터에서 프린터 이름을 변경했을 경우 드라이버 명으로 검색합니다)
                    try
                    {
                        Console.WriteLine(string.Format("Failed to get printer status from pre-printer (Err : {0})", e1.Message));

                        // 설치되어 있는 모든 프린터를 찾습니다
                        using (var searcher = new ManagementObjectSearcher("SELECT * FROM Win32_Printer"))
                        {
                            foreach (var managementObject in searcher.Get())
                            {
                                using (managementObject)
                                {
                                    // 프린터의 드라이버명을 가져옵니다
                                    var name = managementObject["DriverName"].ToString();

                                    foreach (var keyWord in ListKeyWord) {
                                        // 프린터 드라이버명에 키워드가 있는지 찾습니다 (대소문자 구별하지 않고 찾습니다)
                                        if (name.Equals(keyWord, StringComparison.OrdinalIgnoreCase))
                                        {
                                            PrinterName = managementObject["Name"].ToString(); // 프린터를 찾았으면 해당 프린터명을 저장해줍니다
                                            isReady = !Convert.ToBoolean(managementObject["WorkOffline"]);
                                            break;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    catch (Exception e2)
                    {
                        Console.WriteLine(string.Format("Failed to get printer status. Reinstall printer driver (Err : {0})", e2.Message)); // Log
                    }
                }
                return isReady;
            }
        }

        /// <summary>
        /// 인쇄를 시작합니다
        /// </summary>
        /// <param name="frameworkElement"> 인쇄할 Control 객체 </param>
        public void Print(FrameworkElement frameworkElement, double actualWidth, double actualHeight, PrinterPaperModel.PaperType paperType)
        {
            // 프린터가 인쇄할 수 있을 때
            if (IsReady)
            {
                SetPrinter();
                // 드라이버의 정의된 PaperSize 로 사용하려면 SetPrinterPaperSize() API 를 사용하고,
                // 출력하려는 데이터 크기에 따라 CustomPaperSize 를 사용하려면 SetPrinterCustomPaperSize() API 를 사용한다.
                //SetPrinterPaperSize(paperType);
                SetPrinterCustomPaperSize(actualWidth, actualHeight);

                //var fixedDocument = CreateFixedDocument(frameworkElement, paperType);
                var fixedDocument = CreateFixedDocument(frameworkElement, actualWidth, actualHeight);
                _printQueue.CurrentJobSettings.Description = @"Sample Print";
                var xpsDocumentWriter = PrintQueue.CreateXpsDocumentWriter(_printQueue);

                xpsDocumentWriter.Write(fixedDocument, _printQueue.UserPrintTicket);
            }
            else
            {
                Console.WriteLine("Check Printer Status"); // Log
            }
        }

        /// <summary>
        /// 인쇄할 때 사용될 FixedDocument 를 생성합니다
        /// </summary>
        /// <param name="frameworkElement"> 인쇄할 Control 객체 </param>
        /// <param name="paperType"> 인쇄할 용지 타입 </param>
        private FixedDocument CreateFixedDocument(FrameworkElement frameworkElement, PrinterPaperModel.PaperType paperType)
        {
            var fixedDocument = new FixedDocument();
            fixedDocument.DocumentPaginator.PageSize = PrinterPaperModel.GetPrintAreaSize(PrinterPaperModel.PrintableArea[paperType]);

            // 페이지 생성
            var page = new FixedPage
            {
                Background = Brushes.Transparent,
                Margin = new Thickness(0),
                Width = fixedDocument.DocumentPaginator.PageSize.Width,
                Height = fixedDocument.DocumentPaginator.PageSize.Height
            };

            // Control 크기가 인쇄 영역보다 클 경우 Scale 을 조정해주기 위해 아래의 코드를 사용해주세요
            var viewbox = new Viewbox
            {
                Stretch = Stretch.Uniform,
                VerticalAlignment = VerticalAlignment.Top,
                HorizontalAlignment = HorizontalAlignment.Left,
                Width = fixedDocument.DocumentPaginator.PageSize.Width,
                Height = fixedDocument.DocumentPaginator.PageSize.Height,
                Child = frameworkElement,
            };
            page.Children.Add(viewbox);

            // 만약 Control 크기가 인쇄 영역보다 크지 않으면 아래의 코드를 사용해주세요
            // page.Children.Add(frameworkElement);

            // 최종 인쇄 페이지 생성
            var pageContent = new PageContent();
            ((IAddChild)pageContent).AddChild(page);

            fixedDocument.Pages.Add(pageContent);

            return fixedDocument;
        }

        /// <summary>
        /// 입력한 용지 크기로, 인쇄할 때 사용될 FixedDocument 를 생성합니다
        /// </summary>
        /// <param name="frameworkElement"> 인쇄할 Control 객체 </param>
        /// <param name="actualWidth"> 출력 데이터의 Width 값(80mm(3inch) 의 기준 값이 됨) </param>
        /// <param name="actualHeight"> 출력 데이터의 Height 값 </param>
        private FixedDocument CreateFixedDocument(FrameworkElement frameworkElement, double actualWidth, double actualHeight)
        {
            var fixedDocument = new FixedDocument();
            fixedDocument.DocumentPaginator.PageSize = PrinterPaperModel.GetPrintAreaSize(new Size(Calculator.MmToInch(72), Calculator.MmToInch(actualHeight * 72 / actualWidth)));

            var page = new FixedPage
            {
                Background = Brushes.Transparent,
                Margin = new Thickness(0),
                Width = fixedDocument.DocumentPaginator.PageSize.Width,
                Height = fixedDocument.DocumentPaginator.PageSize.Height
            };

            var viewbox = new Viewbox
            {
                Stretch = Stretch.Uniform,
                VerticalAlignment = VerticalAlignment.Top,
                HorizontalAlignment = HorizontalAlignment.Left,
                Width = fixedDocument.DocumentPaginator.PageSize.Width,
                Height = fixedDocument.DocumentPaginator.PageSize.Height,
                Child = frameworkElement,
            };
            page.Children.Add(viewbox);

            var pageContent = new PageContent();
            ((IAddChild)pageContent).AddChild(page);

            fixedDocument.Pages.Add(pageContent);

            return fixedDocument;
        }

        /// <summary>
        /// 인쇄를 수행할 프린터를 선택합니다
        /// </summary>
        private void SetPrinter()
        { 
            // 제어판에 있는 장치 및 프린터를 의미합니다
            using (var localPrintServer = new LocalPrintServer())
            {
                _printQueue = new PrintQueue(localPrintServer, PrinterName);
            }

            // 프린터의 정의된 용지 크기를 가져오기 위해 프린터명으로 객체를 생성합니다
            _printerSettings = new PrinterSettings
            {
                PrinterName = this.PrinterName,
            };
        }

        private void SetSizeToDevMode(byte[] devMode, int index, short value)
        {
            devMode[index] = (byte)(value & 0x00FF);
            devMode[index+1] = (byte)((value & 0xFF00)>>8);
        }

        private void SetPrinterCustomPaperSize(double actualWidth, double actualHeight)
        {

            // 인쇄 정보를 DEVMODE 로 변환합니다
            var printTicketConverter = new PrintTicketConverter(PrinterName, PrintTicketConverter.MaxPrintSchemaVersion);
            var devMode = printTicketConverter.ConvertPrintTicketToDevMode(_printQueue.UserPrintTicket, BaseDevModeType.UserDefault);

            if (actualWidth > actualHeight)
                // dmOrientation
                devMode[76] = (byte)2;

            // dmPaperSize(for custom)
            devMode[78] = (byte)0;

            // dmPaperLength
            var paperLength = (short)(Math.Round(72 * actualHeight / actualWidth) * 10 + 80);
            devMode[80] = (byte)(paperLength & 0x00FF);
            devMode[81] = (byte)((paperLength & 0xFF00) >> 8);

            // dmPaperWidth (fixed value: 800)
            devMode[82] = (byte)(800 & 0x00FF);
            devMode[83] = (byte)((800 & 0xFF00) >> 8);

            _printQueue.UserPrintTicket = printTicketConverter.ConvertDevModeToPrintTicket(devMode);


            // dmPaperLength
            // 드라이버에서는 mm 의 10분의 1 단위의 값으로 용지 사이즈를 규정해서 사용하고 있다.(80mm => 800)
            // 출력하려는 이미지의 높이 값을 계산하는 방법은 다음과 같다.
            // 1. 상/하/좌/우 Margin 이 각 4mm 이고, 너비가 3인치(80mm) 임으로, 실제로 출력되는 너비는 72mm 이다.
            // 2. 출력하려는 이미지의 너비를 72mm 로 잡고, 비율로 높이 값을 계산한다.
            // 3. 드라이버에서는 mm 의 10분의 1 단위의 값으로 용지 사이즈를 사용하기에 mm 로 구한 값에 10배를 하고, Margin 값 8mm(상하)의 10배를 더한다.
            SetSizeToDevMode(devMode, 80, (short)(Math.Round(72 * actualHeight/actualWidth) * 10 + 80));
            // dmPaperWidth(devMode[82] = 0x20, devMode[83] = 0x03 이렇게 값이 들어가야됨.)
            SetSizeToDevMode(devMode, 82, 800);

            _printQueue.UserPrintTicket = printTicketConverter.ConvertDevModeToPrintTicket(devMode);
        }

        /// <summary>
        /// 용지 타입을 보고 인쇄 용지 크기를 설정합니다
        /// (Windows API 인 DEVMODE 를 사용합니다)
        /// (https://docs.microsoft.com/ko-kr/windows/win32/api/wingdi/ns-wingdi-devmodew)
        /// </summary>
        /// <param name="paperType"> 인쇄할 용지 타입 </param>
        private void SetPrinterPaperSize(PrinterPaperModel.PaperType paperType)
        {
            var driverPaperIndex = 3; // Default 80x80
            var paperSize = PrinterPaperModel.PrintableArea[paperType]; // 찾을 용지 타입의 크기를 가져옵니다

            for (int index = 0; index < _printerSettings.PaperSizes.Count; index++)
            {
                int height = 0; // 프린터로부터 받아온 용지 크기의 Height 를 저장할 변수

                // 프린터에서 넘어오는 크기는 용지크기(mm) 를 Inch 로 바꾼 후 * 100 을 한 결과값이 넘어옵니다
                /* 80x80보다 작은 사이즈의 용지 사이즈는 높이를 Width로 가져옵니다.
                    ex ) 80x28 : printerSettings.PaperSizes[index].Width == 79, printerSettings.PaperSizes[index].Height == 283
                          80x56 : printerSettings.PaperSizes[index].Width == 189, printerSettings.PaperSizes[index].Height == 283
                          80x80 : printerSettings.PaperSizes[index].Width == 283, printerSettings.PaperSizes[index].Height == 283
                          80x104 : printerSettings.PaperSizes[index].Width == 283, printerSettings.PaperSizes[index].Height == 378
                          80x136 : printerSettings.PaperSizes[index].Width == 283, printerSettings.PaperSizes[index].Height == 504
                */

                // 프린터 용지 높이 저장
                if (_printerSettings.PaperSizes[index].Width != Calculator.InchToHundred(paperSize.Width))
                    height = _printerSettings.PaperSizes[index].Width;
                else
                    height = _printerSettings.PaperSizes[index].Height;

                // 프린터에 정의되어 있는 용지 Height 값에 인쇄할 용지 타입의 Height 를 뺐을 때 0 이면 인쇄할 용지 크기를 찾은 것입니다
                int result = height - Calculator.InchToHundred(paperSize.Height);

                // 인쇄할 용지 크기를 찾았으면
                if (result == 0)
                {
                    // Driver 는 Index 가 1 부터 시작하기에 찾은 Index 위치에서 + 1 을 해줍니다
                    driverPaperIndex = index + 1;
                    break;
                }
            }

            // 인쇄 정보를 DEVMODE 로 변환합니다
            var printTicketConverter = new PrintTicketConverter(PrinterName, PrintTicketConverter.MaxPrintSchemaVersion);
            var devMode = printTicketConverter.ConvertPrintTicketToDevMode(_printQueue.UserPrintTicket, BaseDevModeType.UserDefault);

            // 인쇄 용지 정보 변경
            devMode[78] = (byte)driverPaperIndex; // 78번 Index 에 인쇄 용지 정보가 들어있기에 78번 Index 를 바꿔줍니다
            Console.WriteLine("Set PaperSize as => " + _printerSettings.PaperSizes[driverPaperIndex - 1]); // Log
            
            _printQueue.UserPrintTicket = printTicketConverter.ConvertDevModeToPrintTicket(devMode);
        }
    }
}
